/* d_logic.cc
 *$Header: /al/acs/src/RCS/d_logic.cc,v 7.98 95/04/19 01:05:14 al Exp $
 * logic model.
 */
#include "ecah.h"
#include "ap.h"
#include "error.h"
#include "u_option.h"
#include "io.h"
#include "u_status.h"
#include "d_logic.h"
#include "s__.h"
#include "declare.h"
/*--------------------------------------------------------------------------*/
//		MODEL_LOGIC::MODEL_LOGIC(const char*);
//	void	MODEL_LOGIC::parse(CS& cmd);
// 	void	MODEL_LOGIC::print(int,int)const;

//		DEV_LOGIC::DEV_LOGIC();
//		DEV_LOGIC::DEV_LOGIC(const DEV_LOGIC& p);
//	void	DEV_LOGIC::parse(CS& cmd);
// 	void	DEV_LOGIC::print(int,int)const;
//	void	DEV_LOGIC::expand();
//	double	DEV_LOGIC::probe_tr_num(const char *what)const;
//	double	DEV_LOGIC::probe_ac_num(const char *what)const;
// 	int	DEV_LOGIC::dotr();
// 	void	DEV_LOGIC::untr();
// 	void	DEV_LOGIC::doac();
static	struct nodestat *tologic(const struct lmod*,const node_t*);
static	double	toanalog(const struct lmod*,const node_t*);
//	double	DEV_LOGIC::tr_review();
/*--------------------------------------------------------------------------*/
extern struct nodestat *nstat;	/* node status flags			    */
static const char *(typename[]) = {"error", "and", "nand", "or", "nor", "xor",
    "inv", ""};
static struct logic defalt = {(generic_t*)NULL, sizeof(struct logic),
    (struct lmod*)NULL, lDEFAULT_modelname, /*more*/};
static struct lmod defaltmodel = {(generic_t*)NULL, sizeof(struct lmod),
    lDEFAULT_delay, lDEFAULT_rise, lDEFAULT_fall, lDEFAULT_rs,
    lDEFAULT_rw, lDEFAULT_th1, lDEFAULT_th0, lDEFAULT_mr, lDEFAULT_mf,
    lDEFAULT_over, lDEFAULT_vmax, lDEFAULT_vmin, lDEFAULT_range};
static MODEL_LOGIC modellist(lDEFAULT_modelname);
/*--------------------------------------------------------------------------*/
MODEL_LOGIC::MODEL_LOGIC(const char *name)
{
  x = create_extra_stuff((generic_t*)&defaltmodel);
  devclass = NOTDEVICE;
  strncpy(label, name, LABELEN);
  label[LABELEN] = '\0';
  stprev = &modellist;
}
/*--------------------------------------------------------------------------*/
void MODEL_LOGIC::parse(CS& cmd)
{
  struct lmod *m;
  m = (struct lmod*)x;
  
  cmd.skiparg();	/* skip known ".model" */
  cmd.ctostr(label, LABELEN, TOKENTERM);
  cmd.skiparg();	/* skip known "logic" */
  cmd.skiplparen();
  for (;;){
    if (cmd.argparse(REPEAT,
	"DElay",aUDOUBLE,	&m->delay,
	"RIse",	aUDOUBLE,	&m->rise,
	"FAll",	aUDOUBLE,	&m->fall,
	"RS",	aUDOUBLE,	&m->rs,
	"RW",	aUDOUBLE,	&m->rw,
	"THH",	aDOUBLE,	&m->th1,
	"THL",	aDOUBLE,	&m->th0,
	"MR",	aUDOUBLE,	&m->mr,
	"MF",	aUDOUBLE,	&m->mf,
	"OVer",	aUDOUBLE,	&m->over,
	"VMAx",	aDOUBLE,	&m->vmax,
	"VMIn",	aDOUBLE,	&m->vmin,
	""))
	;
    else{
      cmd.skiprparen();
      cmd.check(bWARNING);
      break;
    }
    m->range = m->vmax - m->vmin;
  }
}
/*--------------------------------------------------------------------------*/
void MODEL_LOGIC::print(int where, int detail)const
{
  struct lmod *m;
  m = (struct lmod*)x;
  mprintf(where, ".model  %s logic (", label); 
  mprintf(where, " delay=%s ",ftos(m->delay,"", 7, 0));
  mprintf(where, " rise=%s ", ftos(m->rise, "", 7, 0));
  mprintf(where, " fall=%s ", ftos(m->fall, "", 7, 0));
  mprintf(where, " rs=%s ",   ftos(m->rs,   "", 7, 0));
  mprintf(where, " rw=%s ",   ftos(m->rw,   "", 7, 0));
  mprintf(where, " thh=%s ",  ftos(m->th1,  "", 7, 0));
  mprintf(where, " thl=%s ",  ftos(m->th0,  "", 7, 0));
  mprintf(where, " mr=%s ",   ftos(m->mf,   "", 7, 0));
  mprintf(where, " mf=%s ",   ftos(m->mr,   "", 7, 0));
  mprintf(where, " over=%s ", ftos(m->over, "", 7, 0));
  mprintf(where, " vmax=%s ", ftos(m->vmax, "", 7, 0));
  mprintf(where, " vmin=%s ", ftos(m->vmin, "", 7, 0));
  mprintf(where, ")\n");
}
/*--------------------------------------------------------------------------*/
DEV_LOGIC::DEV_LOGIC()
{
  struct logic *xx;
  x = create_extra_stuff((generic_t*)&defalt);
  xx = (struct logic*)x;
  n = xx->n;
  devclass = OTHERDEVICE;
}
/*--------------------------------------------------------------------------*/
DEV_LOGIC::DEV_LOGIC(const DEV_LOGIC& p):ELEMENT(p)
{
  struct logic *xx;
  xx = (struct logic*)x;
  n = xx->n;
}
/*--------------------------------------------------------------------------*/
void DEV_LOGIC::parse(CS& cmd)
{
  struct logic *xx;
  xx = (struct logic*)x;

  parselabel(cmd);
  xx->incount = parsenodes(cmd, PORTSPERGATE);
  cmd.ctostr(xx->modelname, LABELEN, TOKENTERM);
  if (cmd.argparse(ONEPASS,
	"AND",	aENUM,	&xx->type,  lAND,
	"NAND",	aENUM,	&xx->type,  lNAND,
	"OR",	aENUM,	&xx->type,  lOR,
	"NOR",	aENUM,	&xx->type,  lNOR,
	"XOR",	aENUM,	&xx->type,  lXOR,
	"INV",	aENUM,	&xx->type,  lINV,
	""))
    xx->incount -= 2;
  else{
    cmd.check(bWARNING);
  }
}
/*--------------------------------------------------------------------------*/
void DEV_LOGIC::print(int where, int detail)const
{
  struct logic *xx;
  xx = (struct logic*)x;

  printlabel(where);
  printnodes(where);
  mprintf(where, " %s %s\n", xx->modelname, typename[xx->type]); 
}
/*--------------------------------------------------------------------------*/
void DEV_LOGIC::expand()
{
  char cktname[BUFLEN];
  
  struct logic *xx;
  xx = (struct logic*)x;
  assert(xx);
  
  attach_model(&modellist);
  
  sprintf(cktname, "%s%s%u", xx->modelname, typename[xx->type], xx->incount);
  expandsubckt(this,cktname);
  if (!subckt){
    error(bDANGER, "%s: no model, forcing digital\n", printlabel());
  }
  tracesubckt = YES;
  subckt->expand_group();
}
/*--------------------------------------------------------------------------*/
double DEV_LOGIC::probe_tr_num(const char *what)const
{
  CS cmd(what);
  
  node_t ground;
  ground.e = ground.m = ground.t = 0;
  
  if (cmd.pmatch("V")){
    return tr_volts(&(n[OUT2]),&ground);
  }else { /* bad parameter */
    return NOT_VALID;
  }
}
/*--------------------------------------------------------------------------*/
double DEV_LOGIC::probe_ac_num(const char *what)const
{
  return NOT_VALID;
}
/*--------------------------------------------------------------------------*/
int DEV_LOGIC::dotr()
{
  struct nodestat *nn;
  struct logic *xx;
  const struct lmod *mm;
  
  xx = (struct logic*)x;
  mm = xx->m;
  nn = &(nstat[n[1].m]);
  
  if (subckt  &&  ((xx->gatemode == mANALOG)
		   ||  (OPT::mode == mMIXED && SIM::phase == pINIT_DC)
		   ||  (OPT::mode == mANALOG))){
    xx->gatemode = nn->nodemode = mANALOG;
    return converged = subckt->dotr_group();
  }
  
/* if it gets here, either OPT::mode == mMIXED  &&  quality == qGOOD
 *			or OPT::mode == mDIGITAL
 */
  
  bypass = YES;
  xx->gatemode = nn->nodemode = mDIGITAL;
  
  if (SIM::phase == pINIT_DC){
    tr_review();
    nn->finaltime = 0.;
  }
  
  if (SIM::time0 >= nn->finaltime){	/* in transition, time to propagate */
    bypass = NO;
    nn->lv0 = nn->lv1;
    nn->ls0 = nn->ls1;
    nn->quality = qGOOD;
    nn->diter = STATUS::iter[iTOTAL];
    nn->finaltime = BIGBIG;
    nn->lastchange = SIM::time0;
  }
  
  if (nn->finaltime != BIGBIG  ||  !bypass  ||  !SIM::inc_mode){
    node_t vdd;
    node_t ground;
    vdd = n[OUT1];
    ground.e = ground.m = ground.t = 0;
    n[OUT1] = ground;
    trsetup();
    m0.x = 0.;
    y0.x = 0.;
    y0.f1 = -toanalog(mm,&(n[1]));
    y0.f0 = 0.;
    m0.f1 = 1./mm->rs;
    m0.c0 = y0.f1 / -mm->rs;
    trload_passive();
    n[OUT1] = vdd;
  }
  return converged = YES;
}
/*--------------------------------------------------------------------------*/
void DEV_LOGIC::untr()
{
  subckt->untr_group();
  trunload_passive();
}
/*--------------------------------------------------------------------------*/
void DEV_LOGIC::doac()
{
  error(bWARNING, "%s: no logic in AC analysis\n", printlabel());
}
/*--------------------------------------------------------------------------*/
static struct nodestat *tologic(const struct lmod *m, const node_t *node)
{
  struct nodestat *n;
  
  if (node->m == INVALIDNODE){
    node_t ground;
    ground.e = ground.m = ground.t = 0;
    error(bDANGER, "%u:(%u,%u,%u):internal error: invalid node\n",
	  STATUS::iter[iTOTAL], node->e, node->t, node->m);
    return &(nstat[ground.m]);
  }
  
  n = &(nstat[node->m]);
  if (!m){
    m = n->family;
  }else if (!n->family){
    n->family = m;
  }
  
  if (n->nodemode == mDIGITAL){
    ;/* nothing */
  }else{ /* n->nodemode == mANALOG */
    if (n->diter < n->aiter){
      node_t ground;    ground.e = ground.m = ground.t = 0;
      double newv     = tr_volts(node,&ground);
      double sv       = newv / m->range;
      int oldstate    = n->lv0;

      n->dt = SIM::time0 - n->lastchange;
      
      if (sv >= m->th1){
	n->lv0 = n->lv1 = YES;
      }else if (sv <= m->th0){
	n->lv0 = n->lv1 = NO;
      }else{					/* transition region */
	int lv1old   = n->lv1;
	double oldv  = tr_volts_t1(node,&ground);
	double oldsv = oldv / m->range;
	double diff  = sv - oldsv;
	
	if (diff > OPT::abstol){			    /* rising */
	  n->lv1 = YES;
	  if ((n->lv1 == lv1old)
	      &&  (n->lv0 == n->lv1  ||  diff < n->dt/(m->mr * m->rise))){
	    n->quality = qBAD;		/* inflection or too slow */
	    n->failuremode = "slow rise";
	  }
	}else if (diff < -OPT::abstol){		    /* falling */
	  n->lv1 = NO;
	  if ((n->lv1 == lv1old)
	      &&  (n->lv0 == n->lv1  ||  -diff < n->dt/(m->mf * m->fall))){
	    n->quality = qBAD;		/* inflection or too slow */
	    n->failuremode = "slow fall";
	  }
	}
      }
      if (sv > 1.+m->over  ||  sv < -m->over){	/* out of range */
	n->quality = qBAD;
	n->failuremode = "out of range";
      }
      if (n->quality != qGOOD  &&  n->lv0 != oldstate){
	++n->quality;
      }
      n->family = m;
      n->diter = STATUS::iter[iTOTAL];
      n->lastchange = SIM::time0;
    }
  }
  if (m != n->family){
    n->quality = qBAD;
    n->failuremode = "family mismatch";
  } 
  return n;
}
/*--------------------------------------------------------------------------*/
static double toanalog(const struct lmod *m, const node_t *node)
{
  struct nodestat *n;

  n = &(nstat[node->m]);
  if (!m){
    m = n->family;
  }else if (!n->family){
    n->family = m;
  }
  
  if (n->lv0 == n->lv1){
    return (n->lv0) ? m->vmax : m->vmin;
  }else if (SIM::time0 <= (n->finaltime-((n->lv0)?(m->fall):(m->rise)))){
    return (n->lv0) ? m->vmax : m->vmin;
  }else if (SIM::time0 >= n->finaltime){    
    return (n->lv1) ? m->vmax : m->vmin;
  }else{
    double start,end,interp;
    start = (n->lv0) ? m->vmax : m->vmin;
    end   = (n->lv1) ? m->vmax : m->vmin;
    interp = (n->finaltime-SIM::time0) / ((n->lv0)?(m->fall):(m->rise));
    return end - (end - start) * interp;
  }
}
/*--------------------------------------------------------------------------*/
double DEV_LOGIC::tr_review()
{
  struct nodestat *ns[PORTSPERGATE];
  struct logic *xx;
  const struct lmod *m;
  int count;			/* review */
  int lastchangenode = 0;	/* review */
  int lastchangeiter = 0;	/* review, once */
  int quality = qGOOD;		/* both */
  const char* failuremode = "default";	/* review & print */
  
  xx = (struct logic*)x;
  m = xx->m;
  
  for (count = 1;
       count < PORTSPERGATE  &&  this->n[count].e != INVALIDNODE;
       count++){
    ns[count] = tologic(m,&(this->n[count]));
    if (quality > ns[count]->quality){
      quality = ns[count]->quality;
      failuremode = ns[count]->failuremode;
    }
    if (ns[count]->diter >= lastchangeiter){
      lastchangeiter = ns[count]->diter;
      lastchangenode = count;
    }
  }
  
  /* count == number of connections, now const */
  /* if lastchangenode == 1, no new changes, bypass */
  
  if (subckt  &&  
      ((OPT::mode == mANALOG)  ||  (OPT::mode == mMIXED && quality != qGOOD))){
    if (xx->gatemode == mDIGITAL){
      error(bTRACE, "%s:%u:%g switch to analog (review), %s\n",
	    printlabel(), STATUS::iter[iTOTAL], SIM::time0, failuremode);
      trunload_passive();
    }
    xx->gatemode = nstat[this->n[1].m].nodemode = mANALOG;
    return timef = subckt->tr_review_group();
  }
  
  /* if it gets here, either OPT::mode == mMIXED  &&  quality == qGOOD
   *			or OPT::mode == mDIGITAL
   */
  
  if (xx->gatemode == mANALOG){
    error(bTRACE, "%s:%u:%g switch to digital (review)\n",
	  printlabel(), STATUS::iter[iTOTAL], SIM::time0);
    subckt->untr_group();
  }
  xx->gatemode = ns[1]->nodemode = mDIGITAL;
  bypass = YES;
  
  if (!SIM::bypass_ok || lastchangenode!=1 || SIM::phase==pINIT_DC){
    struct nodestat out;
    int ii;
    out = *ns[1];
    bypass = NO;
    switch (xx->type){
    case lAND:
      out.lv1 = YES;
      for (ii = 2;  ii < count;  ii++)
	out.lv1 &= ns[ii]->lv0;
      break;
    case lNAND:
      out.lv1 = YES;
      for (ii = 2;  ii < count;  ii++)
	out.lv1 &= ns[ii]->lv0;
      out.lv1 = !out.lv1;
      break;
    case lOR:
      out.lv1 = NO;
      for (ii = 2;  ii < count;  ii++)
	out.lv1 |= ns[ii]->lv0;
      break;
    case lNOR:
      out.lv1 = NO;
      for (ii = 2;  ii < count;  ii++)
	out.lv1 |= ns[ii]->lv0;
      out.lv1 = !out.lv1;
      break;
    case lXOR:
      out.lv1 = ns[2]->lv0 ^ ns[3]->lv0;
      break;
    case lINV:
      out.lv1 = !ns[2]->lv0;
      break;
    default:
      error(bWARNING,"%s: %s, bad logic type\n",printlabel(),xx->type);
      break;
    }
    if (out.lv1 != ns[1]->lv1){
      if (out.finaltime != BIGBIG){ 			/* in transition */
	out.quality = qBAD;
	out.failuremode = "race";
      }
      out.diter = STATUS::iter[iTOTAL];
      out.finaltime = SIM::time0 + m->delay;
      error(bTRACE, "%s:%u:%g new event\n",
	    printlabel(), STATUS::iter[iTOTAL], SIM::time0);
      new_event(out.finaltime);
      out.lastchange = SIM::time0;
      *ns[1] = out;
      if (lastchangenode == 1){
	error(bDANGER, "%s:%u:%g non-event state change\n",
	      printlabel(), STATUS::iter[iTOTAL], SIM::time0);
      }
    }else if (lastchangenode != 1){
      error(bTRACE,"%s:%u:%g null transition\n",
	    printlabel(), STATUS::iter[iTOTAL], SIM::time0);
    }else{
      error(bTRACE,"%s:%u:%g null evaluation\n",
	    printlabel(), STATUS::iter[iTOTAL], SIM::time0);
    }
  }
  return timef = BIGBIG;
}
/* ***BUG*** review and advance need to be separated.
 * If a step is rejected by review, the logic values could become corrupt.
 */
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
