/* e_card.h		-*- C++ -*-
 *$Header: /al/acs/src/RCS/e_card.h,v 7.98 95/04/19 01:07:10 al Exp $
 * branch structure type definitions
 * device types (enumeration type?) 
 */
#define E_CARD_H

#ifndef E_BASE_H
#include "e_base.h"
#endif

class CS;
class CMD;

#define NODESPERBRANCH	4
#define OUT1	0
#define OUT2	1
#define	IN1	2
#define	IN2	3
#define INVALIDNODE	-1

struct generic_t {
  generic_t *x;
  size_t  ssize;
  const generic_t *m;
  char    modelname[LABELEN+1];
  char	  args[1];	// a fudge.  sizeof() >= 1
};

struct node_t {
  int m;		// mapped, after reordering
  int t;		// m == nm[t] if properly set up
  int e;		// external, user numbers, t = subs exp
};

class CARD : public CKT_BASE {
private:
  static CARD* rootcard;
public:
  static CARD* putbefore;
protected:
  enum devclass_t {NOTSET=-1,NOTDEVICE=0,ONEPORT,TWOPORT,SOURCE,OTHERDEVICE};
  devclass_t devclass;	// something to classify the devices

private:
  CARD*    nextcard;	// linked list stuff
  CARD*    prevcard;
protected:
  generic_t *x;		// extra stuff
  CARD*    stnext;	// list of same type
  CARD*    stprev;
  CARD*    subckt;	// subckt expansion
  CARD*    parent;
  char     label[LABELEN+1];
public: //fanout, insert, allocate
  node_t*  n;
private:
  node_t   nodes[NODESPERBRANCH+1]; // nodes (0,1:out, 2,3:in)
protected:
  double   val;		// value, for simple parts
  double   loss;	// fixed shunt conductance
  double   ic;		// initial condition
  int	   method_a;	// actual integration method (auto)
  int      method_u;	// method to use for this part per user
  int	   probes;	// number of probes set
private:
  int	   evaliter;	// model eval iteration number
protected:
  double   time0;	// time now (y[012], m[01])
  fpoly1_t y1;		// iteration parameters, 1 iter ago
  fpoly1_t y2;		// iteration parameters, 2 iter ago
  fpoly1_t y0;		// iteration parameters, new
  cpoly1_t m0;		// matrix parameters, new
  complex  acg;		// ac admittance matrix values   
  complex  ev;		// ac effective value (usually real)
  double   acbias;	// dc bias for ac use

  unsigned bypass:1,	// is bypassing now
           nodamp:1,	// don't apply iteration damping
	   converged:1;	// convergence status
public: // used by fanout cmd
  unsigned tracesubckt:1; // trace subckt fanout
//--------------------------------------------------------------------
protected:				// special evaluation functions
	void	(*trfun)(CARD*);
	complex	(*acfun)(CARD*);
//--------------------------------------------------------------------
protected:   				// iterators.  e_card1.cc
	void	expand_group();
	int	dotr_group();
	void	untr_group();
	void	doac_group();
	double	tr_review_group();
//--------------------------------------------------------------------
public:					// interface to the iterators - inline
 static	void	expand_all(){rootcard->expand_group();}
 static	int	dotr_all(){return rootcard->dotr_group();}
 static	void	untr_all(){rootcard->untr_group();}
 static	void	doac_all(){rootcard->doac_group();}
 static	double	tr_review_all(){return rootcard->tr_review_group();}
//--------------------------------------------------------------------
public:					// finding one of them - inline
	CARD*	next()const{return nextcard;}
	CARD*	prev()const{return prevcard;}
	CARD*	sub()const{return subckt;}
 static CARD*	root(){return rootcard;}
 static CARD*	first(){return rootcard->next();}
 static CARD*	last(){return rootcard->prev();}
//--------------------------------------------------------------------
public:					// virtuals. -- the important stuff
 virtual CARD*	clone()const = 0;	// either stubs or pure
 virtual void	parse(CS&) = 0;
 virtual void	print(int,int)const = 0;
 virtual void	expand(){}
 virtual int	dotr(){return YES;}
 virtual void	untr(){}
 virtual void	doac(){}
 virtual double	tr_review(){return BIGBIG;};
//--------------------------------------------------------------------
//					// virtuals defined in base
//virtual char*	   printlabel(int where=0)const;    // not overloaded here
//virtual char*    probe_txt(const char*)const;	    // shown here only
//virtual double   probe_num(const char*)const;	    // to complete the list
//virtual double   probe_tr_num(const char*)const;
//virtual double   probe_ac_num(const char*)const;
//virtual xprobe_t probe_ac_ext(const char*)const;
//--------------------------------------------------------------------
protected:				// create and destroy.  e_card2.cc
		CARD();
		CARD(const CARD&);
public:
 virtual	~CARD();
//--------------------------------------------------------------------
public:					// list management.  e_card3.cc
	CARD*	insertbefore(CARD* before);
	CARD*	insertafter(CARD* after);
private:
	void	unlink();
//--------------------------------------------------------------------
public:					// query functions.  e_card4.cc
	char*	printlabel(int where=0)const;
 static	double	probe(const CARD*,const char*);
protected:
	int	evaluated();
public:
	int	is1port()const{return devclass==ONEPORT;}
	int	is2port()const{return devclass==TWOPORT;}
	int	issource()const{return devclass==SOURCE;}
	int	isdevice()const{return devclass>NOTDEVICE;}
//--------------------------------------------------------------------
public:					// modifiers.  inline
	void	setvalue(double v){val = v;}
	void	setvaluex(double v){val = v; x = NULL;}
//--------------------------------------------------------------------
//	friends because the translation to C++ is incomplete
  friend  CARD*  findbranch(CS&,CARD*,CARD*);
  friend  CARD*  findbranch_samescope(const char*,CARD*);
  friend  const CARD* findbranch_sametype(const char*,const CARD*);
  friend  void   expandsubckt(CARD*,const char*);  // e_subckt.h
  friend  complex acfix(CARD*);    // e_exp.h
  friend  void   trfix(CARD*);
  friend  /*static*/ void  trf_max(const CARD*,double**,cpoly1_t*);  
  friend  class  DC;
  friend  class  CARDSTASH;
  friend  class  PROBELIST;
  friend  class  DEV_CCCS;
  friend  class  DEV_CCVS;
  friend  class  DEV_MUTUAL_L;
  friend  class  DEV_INDUCTANCE;
  friend  class  DEV_DIODE;
  friend  class  DEV_LOGIC;
  friend  class  DEV_MOS;
  friend  class  DEV_SUBCKT;
  friend  class  COMPONENT;
  friend  class  BASE_SUBCKT;
};

#define exists(c)  (c)
generic_t*  create_extra_stuff(const generic_t*);

