/* m_matrix.cc		-*- C++ -*-
 *$Header: /al/acs/src/RCS/m_matrix.cc,v 7.97 95/04/16 18:40:15 al Exp $
 * Sparse matrix package
 * Bump and spike - bordered block diagonal pattern
 */
/* This kluge is because there is something wrong with templates
 * on my system (Next - gcc2.2.2).
 * The next release (after I update my compiler) will require working templates
 */
/*--------------------------------------------------------------------------*/
#ifndef BSMATRIX
/*--------------------------------------------------------------------------*/
//#include <assert.h>
//#include <complex.h>
//#include <string.h>
#include "ecah.h"
#include "u_status.h"
#include "u_option.h"
#include "error.h"
//inline int MAX(int x, int y)
//   {return (x>y) ? x : y;}
/*--------------------------------------------------------------------------*/
#define T double
#define BSMATRIX BSMATRIXdouble
#include "m_matrix.cc"
#undef BSMATRIX
#undef T
/*--------------------------------------------------------------------------*/
#define T complex
#define BSMATRIX BSMATRIXcomplex
#include "m_matrix.cc"
#undef BSMATRIX
#undef T
/*--------------------------------------------------------------------------*/
#else
/*--------------------------------------------------------------------------*/
#include "m_matrix.h"
/*--------------------------------------------------------------------------*/
//template <class T>
void BSMATRIX::uninit()
{
  unallocate();
  delete [] Lownode;
  Lownode = NULL;
}
/*--------------------------------------------------------------------------*/
//template <class T>
void BSMATRIX::init(int s)
{
  Trash = Zero = 0.;
  nzcount = 0;
  Size = s;
  rowptr = colptr = diaptr = NULL;
  space = NULL;
  if (s == 0){
    Lownode = NULL;
  }else{
    Lownode = new int[s+1];
    for (int ii = 0;  ii <= Size;  ++ii)
      Lownode[ii] = ii;
  }
}
/*--------------------------------------------------------------------------*/
/* iwant: indicate that "iwant" to allocate this spot in the matrix
 */
//template <class T>
void BSMATRIX::iwant(int node1, int node2)
{
 assert(node1 <= Size);
 assert(node2 <= Size);
 if (node1 == 0  ||  node2 == 0){	/* node 0 is ground, and doesn't    */
    ;					/* count as a connection	    */
 }else if (node1 < Lownode[node2]){
    Lownode[node2]=node1;
 }else if (node2 < Lownode[node1]){
    Lownode[node1]=node2;
 }
}
/*--------------------------------------------------------------------------*/
//template <class T>
void BSMATRIX::unallocate()
{
  assert (Zero == 0.);
  delete [] rowptr;
  delete [] colptr;
  delete [] diaptr;
  delete [] space;
  rowptr = colptr = diaptr = NULL;
  space = NULL;
}
/*--------------------------------------------------------------------------*/
/* allocate: really get the space to work
 */
//template <class T>
void BSMATRIX::allocate()
{
  nzcount = 0;
  for (int ii = 0;   ii <= Size;   ++ii)
    nzcount += 2 * (ii - Lownode[ii]) + 1;

  rowptr = new T*[Size+1];
  colptr = new T*[Size+1];
  diaptr = new T*[Size+1];
  space  = new T[nzcount];

  T* point = space;
  for (ii = 0;   ii <= Size;   ++ii){
    colptr[ii] = point - Lownode[ii];
    rowptr[ii] = colptr[ii] + 2*ii;
    diaptr[ii] = colptr[ii] + ii;
    point += 2 * (ii - Lownode[ii]) + 1;
  }
}
/*--------------------------------------------------------------------------*/
//template <class T>
T& BSMATRIX::s(int row, int col)
{
  assert(col <= Size);
  assert(row <= Size);
  if (col == row){
    return d(row,col);
  }else if (col > row){			/* above the diagonal */
    if (row == 0)
      return Trash;
    else if (row < Lownode[col])
      return Zero;
    else
      return u(row,col);
  }else{/* if (col < row) */		/* below the diagonal */
    if (col == 0)
      return Trash;
    else if (col < Lownode[row])
      return Zero;
    else
      return l(row,col);
  }
}
/*--------------------------------------------------------------------------*/
//template <class T>
double BSMATRIX::density()
{
  nzcount = 0;
  for (int ii = 0;   ii <= Size;   ++ii)
    nzcount += 2 * (ii - Lownode[ii]) + 1;
  return (Size!=0)?(double)nzcount/(Size*Size):0.;
}
/*--------------------------------------------------------------------------*/
//template <class T>
T& BSMATRIX::SubtractDotProduct(int rr, int cc, int dd)
{
 int kk = MAX(lownode(rr), lownode(cc));
 int len = dd - kk;
 T& dot = m(rr, cc);
 if (len > 0){
    T* row = &(l(rr,kk));
    T* col = &(u(kk,cc));
    /* for (ii = kk;   ii < dd;   ii) */
    for (int ii = 0;   ii < len;   ++ii){
       dot -= row[-ii] * col[ii];
    }
 }
 return dot;
}
/*--------------------------------------------------------------------------*/
//template <class T>
T& BSMATRIX::SubtractDotProduct(int rr, int cc, int dd, T& in)
{
 int kk = MAX(lownode(rr), lownode(cc));
 int len = dd - kk;
 T& dot = m(rr, cc);
 dot = in;
 if (len > 0){
    T* row = &(l(rr,kk));
    T* col = &(u(kk,cc));
    /* for (ii = kk;   ii < dd;   ii) */
    for (int ii = 0;   ii < len;   ++ii){
       dot -= row[-ii] * col[ii];
    }
 }
 return dot;
}
/*--------------------------------------------------------------------------*/
//template <class T>
void BSMATRIX::lu_decomp()
{
 int ii, jj, mm;
 int bn;

 STATUS::lud.start();
 for (mm = 1;   mm <= size();   ++mm){
    bn = lownode(mm);
    if (bn < mm){
       u(bn,mm) /= d(bn,bn);
       for (ii =bn+1;  ii<mm;  ii++){
          /* (m(ii,mm) -= dot(ii,mm,ii)) /= d(ii,ii); */
	  SubtractDotProduct(ii,mm,ii) /= d(ii,ii);
       }
       for (jj = bn+1;  jj<mm;  jj++){
          /* m(mm,jj) -= dot(mm,jj,jj); */
	  SubtractDotProduct(mm,jj,jj);
       }
       /* jj == mm */{
          /* m(mm,mm) -= dot(mm,mm,mm); then test */
	  if (SubtractDotProduct(mm,mm,mm) == 0.){
	     error(bWARNING, "open circuit: node %u\n", mm);
	     d(mm,mm) = OPT::gmin;
	  }
       }
    }else{    /* bn == mm */
       if (d(mm,mm)==0.){
	  d(mm,mm) = OPT::gmin;
       }
    }
 }
 STATUS::lud.stop();
}
/*--------------------------------------------------------------------------*/
void BSMATRIX::fbsub(T* v)
{
 int jj, ii;

 STATUS::back.start();
 for (ii = 1; ii <= size(); ++ii){	    /* forward substitution */
    for ( jj=lownode(ii); jj<ii; ++jj ){
       v[ii] -= l(ii,jj) * v[jj];
    }
    v[ii] /= d(ii,ii);
 }
 for (jj = size(); jj > 1; --jj){	    /* back substitution    */
    for (ii = lownode(jj);   ii < jj;   ++ii){
       v[ii] -= u(ii,jj) * v[jj];
    }
 }
 STATUS::back.stop();
}
/*--------------------------------------------------------------------------*/
#endif
/*--------------------------------------------------------------------------*/
